/*
DISCLAIMER FOR PERSONAL IDENTITY VERIFICATION EXAMPLE CODE
May 2006

This software was developed by the National Institute of Standards and
Technology (NIST), an agency of the Federal Government. Pursuant to
Title 15 United States Code Section 105, works of NIST are not subject
to copyright protection in the United States and are considered to be
in the public domain. Thus, the software may be freely reproduced. 

This software is released by NIST as a service and is expressly provided "AS 
IS." NIST MAKES NO WARRANTY OF ANY KIND, EXPRESS, IMPLIED OR STATUTORY, 
INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTY OF MERCHANTABILITY, 
FITNESS FOR A PARTICULAR PURPOSE, NON-INFRINGEMENT AND DATA ACCURACY. NIST 
DOES NOT REPRESENT OR WARRANT THAT THE OPERATION OF THE SOFTWARE WILL BE 
UNINTERRUPTED OR ERROR-FREE, OR THAT ANY DEFECTS WILL BE CORRECTED. 

NIST does not warrant or make any representations regarding the use of the 
software or the results thereof, including but not limited to the 
correctness, accuracy, reliability or usefulness of the software. By using 
this software or by incorporating this software into another product, you 
agree to hold harmless the United Sates Government for any and all damages or 
liabilities that arise out of such use.
*/

/* file csp.c
 *
 * NIST PIV Demo CSP -- Cryptographic Service Provider
 *
 * Designed and Developed by Electrosoft (www.electrosoft-inc.com) based on requirements
 * from the National Institute of Standards and Technology (NIST).
 *
 * This file codes and documents all the necessary CSP functions.
 *
 * A lot of concepts and various context handling code is this file has been
 * adapted from work of Romain Chantereau <rchantereau@entrouvert.com> (CSP11 code)
 */
#ifndef WIN32_LEAN_AND_MEAN
#define WIN32_LEAN_AND_MEAN
#endif

#undef UNICODE                  // ## Not Yet
#include <windows.h>
#include <wincrypt.h>
#include <cspdk.h>
#include <stdio.h>
#include <stdlib.h>
#include <winscard.h>
#include "csp-debug.h"
#include "cspdefs.h"
#include "BERtlv.h"

#ifdef ISO24727
#include "iso24727api.h"
#else
#include "piv.h"
#endif

#include "csp.h"
#include "csputil.h"

HINSTANCE g_hModule = NULL;

SCARDHANDLE cardHandle = 0;
char cachedPIN[20] = "";

static HANDLE    *grantedContexts = (HANDLE) NULL; /**< Granted contexts.*/
static int      grantedContextsNb = 0;  /**< Number of granted Contexts. */

static HANDLE    *grantedKeys = (HANDLE) NULL; /**< Table of all granted keys handles.*/
static int  grantedKeysNb = 0; /**< Number of actual granted keys.*/


BOOL WINAPI
DllMain(
  HINSTANCE hinstDLL,  // handle to the DLL module
  DWORD fdwReason,     // reason for calling function
  LPVOID lpvReserved)  // reserved
{
    if (fdwReason == DLL_PROCESS_ATTACH)
    {
        DisableThreadLibraryCalls(hinstDLL);
        g_hModule = hinstDLL;

        DEBUG(1, "Library attached\n");
        DEBUG(1, "++++++++++++++++++++++++++++++++++++++++++++++++\n");
    }
	if (fdwReason == DLL_PROCESS_DETACH)
    {
        DEBUG(1, "Library detached\n");
        DEBUG(1, "------------------------------------------------\n");
        closeDebug();
    }
    return TRUE;
}

// ###################################################################################
// A Few utility routines
BOOL validateKeyExchangePin(PROV_CTX *provCtx, char *pin)
{
    return piv_login(provCtx, pin, AT_KEYEXCHANGE, cardHandle);
}

BOOL validateSigPin(PROV_CTX *provCtx, char *pin)
{
    return piv_login(provCtx, pin, AT_SIGNATURE, cardHandle);
}

/** \brief Grants a handler. Straight adapted from CSP11 code*/
BOOL grantHandle(HANDLE **grantedHandles, int *length, HANDLE handle)
{
    int index, i;  /* Position indexes in the handlers list.*/
    HANDLE *localList; /* Local copy of the list. */
    HANDLE *localTemp; /* Local reciepient copy for the list.*/
    HANDLE *newList; /* Address of the new list.*/
    char debug[255];

    /** - NULL pointer given, error.*/
    if((grantedHandles == NULL) || (length == NULL))
    {
        SetLastError(E_INVALIDARG);
        return FALSE;
    }

    /** - If given list lenth < 0, set to 0.*/
    if(*length<0)
    {
        *length = 0;
    }
    /** - Copy the granted handlers list to the local copy.*/
    localList = *grantedHandles;

    /** - Allocate space for the new handle.*/
    sprintf(debug, "+Allocating %d.\n", sizeof(HANDLE) * ((*length)+1));
    DEBUG(5,debug);

    localTemp = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY,
                                 sizeof(HANDLE) * ((*length)+1));
    sprintf(debug,"+Allocated at 0x%x.\n", (unsigned int) localTemp);
    DEBUG(5,debug);
    /** - Remember the new list address before changing it.*/
    newList = localTemp;
    /**  - If cannot allocate, error. */
    if(localTemp == NULL)
    {
        SetLastError(NTE_NO_MEMORY);
        *grantedHandles = localList;
        return FALSE;
    }

    /**  - Copy old list to new list.*/
    for(i=0; i<*length; i++)
    {
        *localTemp = *localList;
        localTemp++;
        localList++;
    }
    /**  - Index is set to the last position.*/
    index = *length;

    /** - If the context index is valid.*/
    if (index>=0 && index<(*length)+1)
    {
        /**  - Fill the new list with the provided context.*/
        newList[index] = handle;
        /**  - Free old list if not NULL.*/
        if(*grantedHandles != NULL)
        {
            sprintf(debug,"+Freeing 0x%x\n", (unsigned int) *grantedHandles);
            DEBUG(5,debug);
            HeapFree(GetProcessHeap(), 0, *grantedHandles);
            sprintf(debug, "+Freed\n");
            DEBUG(5, debug);
        }
        /**  - Copy new list address to the grantedHandles pointer.*/
        *grantedHandles = newList;
        /**  - Now the list is 1 more lenth.*/
        *length=(*length) + 1;
         /**  - Return TRUE.*/
        sprintf(debug,"+(0x%x, %d)\n",(unsigned int) *grantedHandles, *length);
        DEBUG(3, debug);
        return TRUE;
    }
    else
    {
        /** - Else, return FALSE.*/
        return FALSE;
    }
}

/** \brief Revokes a granted cryptographic handler. Straight adapted from CSP11 code*/
BOOL revokeHandle(HANDLE **grantedHandles, int *length, HANDLE handle)
{
    int index, i;  /* Position indexes in the handlers list.*/
    HANDLE *localList; /* Local copy of the list. */
    HANDLE *localTemp; /* Local reciepient copy for the list.*/
    HANDLE *newList; /* Local copy of the new list.*/
    char debug[255];

    /** - NULL pointer given, error.*/
    if((grantedHandles == NULL) || (length == NULL))
    {
        SetLastError(E_INVALIDARG);
        return FALSE;
    }
    sprintf(debug, "-(0x%x, %d)\n",(unsigned int) *grantedHandles, *length);
    DEBUG(3, debug);

    /** - Find the revoked handler position index. */
    index = findGrantedHandle(*grantedHandles, *length, handle);
    sprintf(debug,"-%d to be removed\n", index+1);
    DEBUG(5, debug);

    /** - If length is 0, no possible granted handles.*/
    if (*length <= 0)
    {
        SetLastError(E_INVALIDARG);
        return FALSE;
    }
    /** - If index < 0, no such granted handle.*/
    if (index < 0)
    {
        SetLastError(E_INVALIDARG);
        return FALSE;
    }
    /** - If there is more than one, after remove, it will some left.*/
    if(*length > 1)
    {
        /** - Local granted handlers pointer copy.*/
        localList = *grantedHandles;
        /** - Allocate for length handlers if there are still some.*/
        sprintf(debug, "-Allocating %d.\n", sizeof(HANDLE) * ((*length)-1));
        DEBUG(5, debug);
        localTemp = HeapAlloc(GetProcessHeap(), HEAP_ZERO_MEMORY,
                                 sizeof(HANDLE) * (*length)-1);
        /** - Remember new list address before using it.*/
        newList = localTemp;
        sprintf(debug, "-Allocated at 0x%x.\n", (unsigned int) localTemp);
        DEBUG(5,debug);
        /** - If cannot allocate, error. */
        if(localTemp == NULL)
        {
            SetLastError(NTE_NO_MEMORY);
            *grantedHandles = localList;
            return FALSE;
        }
        /** - Walk list and new list for copy left handles.*/
        for(i=0; i<*length; i++)
        {
            /**  - If i different from revoked handler index, copy.*/
            if(i != index)
            {
                *localTemp = *localList;
                localTemp++;
            }
            localList++;
        }
    }
    /** - Else, there will no more left after revoking.*/
    else
    {
        sprintf(debug,"-No more left.\n");
        DEBUG(5, debug);
        newList = NULL;
    }


    /** - Now the list is 1 less length.*/
    *length=*length - 1;
    /** - Free memory used by the old list.*/
    sprintf(debug,"-Freeing 0x%x\n", (unsigned int) *grantedHandles);
    DEBUG(5, debug);
    HeapFree(GetProcessHeap(), 0, *grantedHandles);
    sprintf(debug, "-Freed.\n");
    DEBUG(5, debug);
    /** - Set the granted handlers pointer to the new list address.*/
    *grantedHandles = newList;
    sprintf(debug,"-(0x%x, %d)\n",(unsigned int)*grantedHandles, *length);
    DEBUG(3, debug);
    return TRUE;
}

/** \brief Return the wanted handler index in the granted handlers list.*/
int findGrantedHandle(HANDLE *grantedHandles, int max_len, HANDLE wantedHandle)
{
    HANDLE *current;    /* Current handler index in the handlers list.*/
    int index;  /* Position index in the handlers list.*/
    BOOL found; /* If the handle is found.*/
    char debug[255];

    /** - Copy the grantedHandles start pointer to a local copy.*/
    current = grantedHandles;
    index = 0;

    /** - If not NULL.*/
    if(current == NULL)
    {
        return -1;
    }

    sprintf(debug,"Search from 0x%x\n",(unsigned int) current);
    DEBUG(5,debug);
    /** - Not found.*/
    found = FALSE;
    /** - Check each entry until we found a empty one or end of list.*/
    while(!found && (index <max_len))
    {
        sprintf(debug,"0x%x =?= 0x%x\n",(unsigned int) *current,
                                        (unsigned int) wantedHandle);
        DEBUG(5,debug);
        /**  - If not the same,next.*/
        if(*current != wantedHandle)
        {
            current++;
            index++;
        }
        /**  - Else found.*/
        else
        {
            found = TRUE;
        }
    }
    /** - If found.*/
    if (found)
    {
        /**  - Return the index position.*/
        return index;
    }
    /** - Else.*/
    else
    {
        /**  - Return negative value.*/
        return -1;
    }
}

BOOL grantedHandle(HANDLE *grantedHandles, int length, HANDLE handle)
{
    /** - If the wanted handle has a position index >=0, it has been granted.*/
    if(findGrantedHandle(grantedHandles, length, handle)>=0)
    {
        return TRUE;
    }
    else
    {
        return FALSE;
    }
}

/*
 -  CPAcquireContext
 -
 *  Purpose:
 *               The CPAcquireContext function is used to acquire a context
 *               handle to a cryptographic service provider (CSP).
 *
 *
 *  Parameters:
 *               OUT phProv         -  Handle to a CSP
 *               IN  szContainer    -  Pointer to a string which is the
 *                                     identity of the logged on user
 *               IN  dwFlags        -  Flags values
 *               IN  pVTable        -  Pointer to table of function pointers
 *
 *  Returns:
 */

BOOL WINAPI
CPAcquireContext(
    OUT HCRYPTPROV *phProv,
    IN  LPCSTR szContainer,
    IN  DWORD dwFlags,
    IN  PVTableProvStruc pVTable)
{
	BOOL ret;
	char debug[255];

	HCRYPTPROV hCryptProv = 0;
	LPCSTR cName = NULL;            /*  Local copy of szContainer C string pointer */
    size_t nameSize = 0;        /* Size of the provided container name.*/
    PROV_CTX *pProvCtx = NULL;      /* Provider context */
    HANDLE  heapHandle;             /* Handle to the context heap object. */
    DWORD FuncReturnedhWnd = 0;

    DEBUG(2,"------------- Acquiring Context ----------------\n");
	sprintf(debug, "param2 value %s\n", szContainer);
	DEBUG(3,debug);
	sprintf(debug, "param3 value %d\n", dwFlags);
	DEBUG(3,debug);

	sprintf(debug, "Calling Thread ID is %d\n", GetCurrentThreadId());
	DEBUG(2,debug);

	sprintf(debug, "Calling process ID is %d\n", GetCurrentProcessId());
	DEBUG(2,debug);

    /** - Nullify the returned hProv.*/
    *phProv = (HCRYPTPROV)NULL;

    /** - Test if dwFlags are correct */
    if (dwFlags & ~(CRYPT_SILENT|CRYPT_MACHINE_KEYSET))
    {
        SetLastError(NTE_BAD_FLAGS);
        return FALSE;
    }

    /** - If szContainer contains string address */
    if (szContainer != NULL)
    {
        /**  - Test if the container name is valid */
        /** \note Cannot use StringCbLength, not supported in Cygwin/minGW*/
        /* if FAILED(StringCbLength(szContainer, MAX_PATH, nameSize))*/
        nameSize = strlen(szContainer);

        /**  - Test if the name is not too long.*/
        if(nameSize >= MAX_PATH)
        {
            SetLastError(NTE_BAD_KEYSET_PARAM);
            return FALSE;
        }
		sprintf(debug, "Container string size value %d\n", nameSize);
		DEBUG(3, debug);
        /**  - Test if the name is not empty.*/
        if(!szContainer[0])
        {
            szContainer = NULL;
        }
    }
    /**  - Copy name into local variable.*/
    cName = szContainer;

    /** - Allocating provider context, initial size of sizeof(PROV_CTX),
     *    growable (limit =0). */
    heapHandle = HeapCreate(0, sizeof(PROV_CTX), 0);
    /** - Fill with zero to be clean.*/
    pProvCtx = HeapAlloc(heapHandle, HEAP_ZERO_MEMORY, sizeof(PROV_CTX));
    /** - Test if allocation succeed.*/
    if (pProvCtx == NULL)
    {
        SetLastError(NTE_NO_MEMORY);
        return FALSE;
    }
    /** - Fill provider context handle pointer by the
    *    pointer to the acquired provider context.*/
    *phProv = (HCRYPTPROV) pProvCtx;
    /** - Grant the context.*/
    DEBUG(3,"===provCtx:");
    grantHandle((HANDLE **) &grantedContexts,
                                    &grantedContextsNb, (HANDLE) *phProv);
    /** - Fill context heap heandler and provider type */
    pProvCtx->heap = heapHandle;
    pProvCtx->dwProvType = pVTable->dwProvType;
    pProvCtx->cachedKeyExchangePin = NULL;
    pProvCtx->cachedSigPin = NULL;
    pProvCtx->nocache = FALSE;
    pProvCtx->silent = FALSE;
    pProvCtx->currentAlg = 0;
    pProvCtx->uiHandle = 0;
    pProvCtx->container.cName = NULL;
    pProvCtx->container.dwFlags = 0;
    pProvCtx->container.dwContainerType = -1;
    pProvCtx->container.hServiceInformation = NULL;

	/** - Transmit the cps11 instance handle to service. We dont need this - Shouvik
    setCSPInstance(g_hModule);
    */

    /** - Fill container flags. */
    pProvCtx->container.dwFlags = dwFlags;
    /** - Test if verify is not set.*/
    /** - Fill container type.*/
    pProvCtx->container.dwContainerType = SC_CONTAINER;
    /** - Test if machine keyset is not set.*/
    if(pProvCtx->container.dwFlags & CRYPT_MACHINE_KEYSET)
    {
        /**  -If set, no data caching.*/
        pProvCtx->nocache = TRUE;
    }
    /** - Test if context have to be silent.*/
    if(pProvCtx->container.dwFlags & CRYPT_SILENT)
    {
        /**  - If set, silent context.*/
        pProvCtx->silent = TRUE;
    }
    /** - Store the wanted key container in the provider contexr. */
    /**  - if name set.*/
    if (cName)
    {
        /**   - Allocate space in the container.*/
        pProvCtx->container.cName = HeapAlloc(pProvCtx->heap,
                                            HEAP_ZERO_MEMORY, nameSize+1);
        /**   - Test if allocation succeed.*/
        if(pProvCtx->container.cName == NULL)
        {
            SetLastError(NTE_NO_MEMORY);
            CPReleaseContext(*phProv,0);
            return FALSE;
        }
        /**   - Copy name in the container.*/
        if (!strncpy(pProvCtx->container.cName, cName, nameSize))
        {
            SetLastError(NTE_BAD_KEYSET_PARAM);
            CPReleaseContext(*phProv,0);
            return FALSE;
        }
		pProvCtx->container.cName[nameSize] = '\0';
    }
    else
    {
        /**  - No name set=> NULL in container name.*/
        /** cName is not set, we will use the defaut container. */
        pProvCtx->container.cName = NULL;
    }

#ifdef ISO24727
	// initialize ISO/IEC 24727-3 layer
	if (Initialize() != CSP_S_OK)
	{
		CPReleaseContext(*phProv,0);
		DEBUG(2, "Failed to initialize ISO/IEC 24727-3 layer\n");
		return FALSE;
	}
#endif

    /** Find the specified container and fill container internal
     * information.*/
    /** - Open the container.*/
    if ( openPIVContainer(pProvCtx, &cardHandle) == FALSE )
    {
        CPReleaseContext(*phProv,0);
		DEBUG(2, "openPIVContainer failed\n");
        return FALSE;
    }

	DEBUG(2, "Going to get after MS Provider for help\n");

	// Ask MS Provider for help!!!
	ret = CryptAcquireContext(
        &hCryptProv,
        NULL,
        NULL,
        PROV_RSA_FULL,
        CRYPT_VERIFYCONTEXT);

	if( !ret )
	{
		sprintf(debug, " CryptAcquireContext failed for helper CSP - Last error ->%x", GetLastError());
		CPReleaseContext(*phProv,0);
		DEBUG(3, debug);
		return FALSE;
	}
	pProvCtx->msProv = hCryptProv;
	DEBUG(2,"--------- Acquired Context ------------\n");
	return TRUE;
}


/*
 -      CPReleaseContext
 -
 *      Purpose:
 *               The CPReleaseContext function is used to release a
 *               context created by CryptAcquireContext.
 *
 *     Parameters:
 *               IN  phProv        -  Handle to a CSP
 *               IN  dwFlags       -  Flags values
 *
 *  Returns:
 */

BOOL WINAPI
CPReleaseContext(
    IN  HCRYPTPROV hProv,
    IN  DWORD dwFlags)
{

	char debug[255];

	HANDLE  heapHandle;             /* Handle to the context heap object. */
    PROV_CTX *pProvCtx = NULL;      /* Provider context */

	sprintf(debug, "hProv value %x\n", hProv);

    DEBUG(2,"---------------------Releasing Context-----------------\n");

	sprintf(debug, "Calling Thread ID is %d\n", GetCurrentThreadId());
	DEBUG(2,debug);

	sprintf(debug, "Calling process ID is %d\n", GetCurrentProcessId());
	DEBUG(2,debug);

    /** - Local good typed provider context pointer.*/
    pProvCtx = (PROV_CTX *) hProv;

    /** - Test if flags are set.*/
    if(dwFlags)
    {
        SetLastError(NTE_BAD_FLAGS);
        return FALSE;
    }

    /** - Test if the context has been granted.*/
    if(!grantedHandle((HANDLE *) grantedContexts, grantedContextsNb, (HANDLE) hProv))
    {
        SetLastError(NTE_BAD_UID);
        return FALSE;
    }
    DEBUG(3,"ReleaseContext - Granted handle...proceed\n");
    /** - If there are dependent context handles, release context.*/
    if(pProvCtx->msProv != 0)
    {
        if(!CryptReleaseContext(pProvCtx->msProv, 0))
        {
            return FALSE;
        }
    }
	if(pProvCtx->msProv2 != 0)
    {
        if(!CryptReleaseContext(pProvCtx->msProv2, 0))
        {
            return FALSE;
        }
		// msProv2 was used to create a temp container...lets delete that
		CryptAcquireContext(&(pProvCtx->msProv2), "TestContainer", MS_ENHANCED_PROV,
				PROV_RSA_FULL, CRYPT_DELETEKEYSET);
    }
	DEBUG(3,"ReleaseContext - Helper CSP release context done...proceed\n");

	if(!closePIVContainer(pProvCtx, &cardHandle))
    {
		DEBUG(2,"ReleaseContext - PIV Container release failed...but proceed\n");
	    //return FALSE;
    }

    DEBUG(2,"ReleaseContext - PIV Container close came back\n");

#ifdef ISO24727
	// terminate access to ISO/IEC 24727-3 layer
	Terminate();
	DEBUG(2,"Terminated access to ISO/IEC 24727-3 layer\n");
#endif

    /** - Revoke the context handle.*/
    heapHandle = pProvCtx->heap;
    DEBUG(3, "===provCtx:");
    if(!revokeHandle((HANDLE **) &grantedContexts, &grantedContextsNb, (HANDLE) hProv))
    {
        return FALSE;
    }
    pProvCtx->heap = NULL;
    if(pProvCtx->cachedKeyExchangePin != NULL)
    {
        HeapFree(heapHandle, 0, pProvCtx->cachedKeyExchangePin);
    }
    if(pProvCtx->cachedSigPin != NULL)
    {
        HeapFree(heapHandle, 0, pProvCtx->cachedSigPin);
    }
    /** - Free context .*/
    HeapFree(heapHandle, 0, pProvCtx);
    /** - Destroy context heap handle.*/
    HeapDestroy(heapHandle);
    DEBUG(2, "-----------------Context released-----------------\n");
    return TRUE;
}

/*
 -  CPDestroyKey
 -
 *  Purpose:
 *                Destroys the cryptographic key that is being referenced
 *                with the hKey parameter
 *
 *
 *  Parameters:
 *               IN      hProv  -  Handle to a CSP
 *               IN      hKey   -  Handle to a key
 *
 *  Returns:
 */

BOOL WINAPI
CPDestroyKey(
    IN  HCRYPTPROV hProv,
    IN  HCRYPTKEY hKey)
{
    /* return TRUE; */
	unsigned char debug[100];
	BOOL ret;

	PROV_CTX *pProvCtx = NULL;      /* Provider context */
    KEY_INFO *pKeyInfo; /* Handle to the generated key.*/

    DEBUG(3,"_-_-_-_-_-_-_-_-_Destroying Key-_-_-_-_-_-_-_-_-\n");
    /** - Local copy of the crypto handler.*/
    pProvCtx = (PROV_CTX *) hProv;

	sprintf(debug, "Incoming key handle for destruction = %x\n", hKey);
	DEBUG(3, debug);

    /** - Local copy of the key handler.*/
    pKeyInfo = (KEY_INFO *) hKey;

    /** - Test if tha handle has been granted.*/
    if(!grantedHandle((HANDLE *) grantedContexts, grantedContextsNb, (HANDLE) hProv))
    {
        SetLastError(NTE_BAD_UID);
        return FALSE;
    }
    /** - Test if key handle has been granted.*/
    if(!grantedHandle((HANDLE *) grantedKeys, grantedKeysNb, (HANDLE) hKey))
    {
		// if we are not managing this key, maybe our helper MS Provider is
		// Lets try there.....
		ret = CryptDestroyKey(hKey); //
		if( !ret )
		{
			DEBUG(3, "MS Provider failed in destroyed key\n");
			SetLastError(NTE_BAD_KEY);
			return FALSE;
		}
		else
		{
			DEBUG(3, "MS Provider destroyed key sucessfully\n");
			return TRUE;
		}
    }

    /** - Destroy them.*/
	// All we need to do is to free the memory gotten by data
	if(pKeyInfo->data)
	{
		free(pKeyInfo->data);
	}

    /** - Revoke handle.*/
    DEBUG(3,"===hKey:");
    if(revokeHandle((HANDLE **) &grantedKeys, &grantedKeysNb, (HANDLE) hKey))
    {
        /** - Free key info.*/
        if(!HeapFree(pProvCtx->heap, 0, pKeyInfo))
        {
			DEBUG(3, "Returning bad from Destroy Key(1) due to NTE_BAD_KEY\n");
            SetLastError(NTE_BAD_KEY);
            return FALSE;
        }
    }
    else
    {
		DEBUG(3, "Returning bad from Destroy Key(2) due to NTE_BAD_KEY\n");
        SetLastError(NTE_BAD_KEY);
        return FALSE;
    }
	DEBUG(3, "Returning fine from Destroy Key\n");
	return TRUE;
}

/*
 -  CPGetKeyParam
 -
 *  Purpose:
 *                Allows applications to get various aspects of the
 *                operations of a key
 *
 *  Parameters:
 *               IN      hProv      -  Handle to a CSP
 *               IN      hKey       -  Handle to a key
 *               IN      dwParam    -  Parameter number
 *               OUT     pbData     -  Pointer to data
 *               IN      pdwDataLen -  Length of parameter data
 *               IN      dwFlags    -  Flags values
 *
 *  Returns:
 */

BOOL WINAPI
CPGetKeyParam(
    IN  HCRYPTPROV hProv,
    IN  HCRYPTKEY hKey,
    IN  DWORD dwParam,
    OUT LPBYTE pbData,
    IN OUT LPDWORD pcbDataLen,
    IN  DWORD dwFlags)
{
	/*
    *pcbDataLen = 0;
    return TRUE;
	*/
	unsigned char debug[4000];
	char base64data[4000];
	int datalen, len, ii;

	PROV_CTX *pProvCtx = NULL;      /* Provider context */
    KEY_INFO *pKeyInfo; /* Handle to the generated key.*/

    DEBUG(3,"_-_-_-_-_-_-_-_-_Getting key parameters-_-_-_-_-_-_-_-_-\n");
	sprintf(debug, "dwParam is %d\n", dwParam);
	DEBUG(3, debug);
    /** - Local copy of the crypto handler.*/
    pProvCtx = (PROV_CTX *) hProv;

    /** - Local copy of the key handler.*/
    pKeyInfo = (KEY_INFO *) hKey;

    /** - Test if tha handle has been granted.*/
    if(!grantedHandle((HANDLE *) grantedContexts, grantedContextsNb, (HANDLE) hProv))
    {
        SetLastError(NTE_BAD_UID);
        return FALSE;
    }

    /** - Test if key handle has been granted.*/
    if(!grantedHandle((HANDLE *) grantedKeys, grantedKeysNb, (HANDLE) hKey))
    {
        SetLastError(NTE_BAD_KEY);
        return FALSE;
    }

	if (dwParam != KP_CERTIFICATE)
    {
        SetLastError(NTE_BAD_FLAGS);
        return FALSE;
    }

    /** - Test Flags validity.*/
    if (dwFlags)
    {
        SetLastError(NTE_BAD_FLAGS);
        return FALSE;
    }

	if(pbData == NULL)
    /** - If first call, retrieve data length.*/
    /**   \warning This size is given in BYTES.*/
    {
		*pcbDataLen = pKeyInfo->dataLen;
	}
	else
	{
		if(*pcbDataLen < pKeyInfo->dataLen)
        {
            SetLastError(ERROR_MORE_DATA);
            return FALSE;
        }
		*pcbDataLen = pKeyInfo->dataLen;
        /**  - Fill with CERT data.*/
        memcpy(pbData, pKeyInfo->data, *pcbDataLen);
	}

	if( pbData != NULL )
	{
		len = *pcbDataLen;
		sprintf(debug, "Data length of cert is -  %d\n", len);
		DEBUG(3, debug);

		datalen = 4000;
		if(Base64Encode(pbData, len, base64data, &datalen, 0))
		{
			DEBUG(3, "Base64Encode encoding fine!!!");
		}
		else
		{
			DEBUG(3, "Base64Encode encoding screwed!!!");
		}
		sprintf(debug, "Total base64 bytes returned ->%d\n", datalen);
		DEBUG(3, debug);
		for(ii=0; ii < datalen; ii++)
		{
			debug[ii] = base64data[ii];
		}
		debug[ii] = '\0';
		DEBUG(3, debug);
		sprintf(debug, "\n\n#################  $$$$$$$$$$$$$$$$$$$$$$$$$$\n");
		DEBUG(3, debug);
	}

	DEBUG(3, "Returning fine from GetKeyParam\n");
	return TRUE;
}


/*
 -  CPSetProvParam
 -
 *  Purpose:
 *                Allows applications to customize various aspects of the
 *                operations of a provider
 *
 *  Parameters:
 *               IN      hProv   -  Handle to a CSP
 *               IN      dwParam -  Parameter number
 *               IN      pbData  -  Pointer to data
 *               IN      dwFlags -  Flags values
 *
 *  Returns:
 */

BOOL WINAPI
CPSetProvParam(
    IN  HCRYPTPROV hProv,
    IN  DWORD dwParam,
    IN  CONST BYTE *pbData,
    IN  DWORD dwFlags)
{
    /*return TRUE;*/
	unsigned char debug[200];
    DWORD   dataLen = 0; /* Returned data lenth in bytes.*/
    char *localPin = NULL; /* The transmitted cached PIN.*/
    PROV_CTX *pProvCtx = NULL; /* The local casted copy of the provider
                                    context pointer.*/

    DEBUG(3,"_-_-_-_-_-_-_-_-_Setting provider parameters-_-_-_-_-_-_-_-_-\n");
    /** - Local copy of the crypto handle.*/
    pProvCtx = (PROV_CTX *) hProv;

	sprintf(debug, "hProv value %x\n", hProv);
	DEBUG(3, debug);

    /** - Test if c handle has been granted.*/
    if(!grantedHandle(grantedContexts, grantedContextsNb, (HANDLE) hProv))
    {
        SetLastError(NTE_BAD_UID);
        return FALSE;
    }

    /** - Test flag existenz.*/
    if(dwFlags)
    {
        SetLastError(NTE_BAD_FLAGS);
        return FALSE;
    }

	DEBUG(3, "Starting to check params etc\n");
    switch(dwParam)
    {
        case PP_KEYEXCHANGE_PIN: /** - PP_KEYEXCHANGE_PIN:*/
            /**  - Copy pbData length to dataLen.*/
            dataLen = strlen(pbData);
            /**  - Test if pbData length <= 8.*/
            if(dataLen>8)
            {
                SetLastError(ERROR_INVALID_PARAMETER);
                return FALSE;
            }
            /**  - Allocate space for localPin.*/
            localPin = HeapAlloc(pProvCtx->heap, HEAP_ZERO_MEMORY, dataLen +1);
            /**  - Copy to local pin.*/
            strncpy(localPin,pbData,dataLen);
			localPin[dataLen] = '\0';
            /**  - If cached PIN already exists.*/
			DEBUG(3, "Processing Exchange Pin values\n");
            if(pProvCtx->cachedKeyExchangePin != NULL)
            {
				DEBUG(3, "Here 111\n");
                /**  - Test if provided and cached PIN are the same.*/
                if(strcmp(localPin, pProvCtx->cachedKeyExchangePin))
                {
                    SetLastError(SCARD_W_WRONG_CHV);
                    return FALSE;
                }
                /** \todo Test logonID, but what is it - we dont care for now Shouvik!!!*/
            }
            /**  - Else, validate pin.*/
            else
            {
				DEBUG(3, "Here 222\n");
                if(validateKeyExchangePin(pProvCtx, localPin))
                {
					DEBUG(3, "validateKeyExchangePin came back good\n");
                    pProvCtx->cachedKeyExchangePin = localPin;
					// Copy data to cached PIN
					strcpy(cachedPIN, localPin);
					sprintf(debug, "Pin should be cached now -> %s\n", cachedPIN);
					DEBUG(3, debug);
                }
                else
                {
					DEBUG(3, "validateKeyExchangePin failed\n");
                    SetLastError(SCARD_W_WRONG_CHV);
                    return FALSE;
                }
            }
            break;
        case PP_SIGNATURE_PIN: /**  - PP_SIGNATURE_PIN:*/
            /**  - Copy pbData length to dataLen.*/
            dataLen = strlen(pbData);
            /**  - Test if pbData length <= 8.*/
            if(dataLen>8)
            {
                SetLastError(ERROR_INVALID_PARAMETER);
                return FALSE;
            }
            /**  - Allocate space for localPin.*/
            localPin = HeapAlloc(pProvCtx->heap, HEAP_ZERO_MEMORY, dataLen +1);
            /**  - Copy to local pin.*/
            strncpy(localPin,pbData,dataLen);
			localPin[dataLen] = '\0';
            /**  - If cached PIN already exists.*/
            if(pProvCtx->cachedSigPin != NULL)
            {
                /**  - Test if provided and cached PIN are the same.*/
                if(strcmp(localPin, pProvCtx->cachedSigPin))
                {
                    SetLastError(SCARD_W_WRONG_CHV);
                    return FALSE;
                }
            }
            /**  - Else, validate pin.*/
            else
            {
                if(validateSigPin(pProvCtx, localPin))
                {
                    pProvCtx->cachedKeyExchangePin = localPin;
                }
                else
                {
                    SetLastError(SCARD_W_WRONG_CHV);
                    return FALSE;
                }
            }
            break;
        default:
			DEBUG(3, "Bad param found in SetProvParam\n");
            SetLastError(NTE_BAD_TYPE);
            return FALSE;
    }
	DEBUG(3, "SetProvParam returning TRUE\n");
    return TRUE;
}


/*
 -  CPGetProvParam
 -
 *  Purpose:
 *                Allows applications to get various aspects of the
 *                operations of a provider
 *
 *  Parameters:
 *               IN      hProv      -  Handle to a CSP
 *               IN      dwParam    -  Parameter number
 *               OUT     pbData     -  Pointer to data
 *               IN OUT  pdwDataLen -  Length of parameter data
 *               IN      dwFlags    -  Flags values
 *
 *  Returns:
 */

BOOL WINAPI
CPGetProvParam(
    IN  HCRYPTPROV hProv,
    IN  DWORD dwParam,
    OUT LPBYTE pbData,
    IN OUT LPDWORD pcbDataLen,
    IN  DWORD dwFlags)
{
    /*pcbDataLen = 0;return TRUE;*/
	unsigned char debug[100];

	DWORD   dataLen = 0; /* Returned data lenth in bytes.*/
    PROV_ENUMALGS *enumAlg = NULL; /* The enumerated Alg.*/
    PROV_ENUMALGS_EX *enumAlgEx = NULL; /* The enumerated Alg.*/
    char *localPin = NULL; /* The transmitted cached PIN.*/
    PROV_CTX *pProvCtx = NULL; /* The local casted copy of the provider
                                    context pointer.*/
    ALGORITHM algorithm;    /* Local algorithm information.*/

    DEBUG(3,"_-_-_-_-_-_-_-_-_Getting provider parameters-_-_-_-_-_-_-_-_-\n");
	sprintf(debug, "param dwparam is %d\n", dwParam);
	DEBUG(3, debug);
	sprintf(debug, "param pbData is %x\n", pbData);
	DEBUG(3, debug);
	sprintf(debug, "param pcbDataLen is %d\n", *pcbDataLen);
	DEBUG(3, debug);
	sprintf(debug, "param dwFlags is %d\n", dwFlags);
	DEBUG(3, debug);
    /** - Local copy of the crypto handle.*/
    pProvCtx = (PROV_CTX *) hProv;

    /** - Test if c handle has been granted.*/
    if(!grantedHandle(grantedContexts, grantedContextsNb, (HANDLE) hProv))
    {
		DEBUG(3, "Returning NTE_BAD_UID from GetProvParam\n");
        SetLastError(NTE_BAD_UID);
        return FALSE;
    }

    /** - Test flag colours.*/
    if(dwFlags && (dwFlags & ~(CRYPT_FIRST)))
    {
		DEBUG(3, "Returning NTE_BAD_FLAGS from GetProvParam\n");
        SetLastError(NTE_BAD_FLAGS);
        return FALSE;

    }
    switch(dwParam)
    {
        case PP_CONTAINER: /**  - PP_CONTAINER:*/
            /**   - If container name is NULL.
            if(pProvCtx->container.cName == NULL)
            {
                HCRYPTKEY hKey;
            }
			*/
            /**   - If container name is not NULL.*/
            if(pProvCtx->container.cName != NULL)
            {
                /**    - dataLen is the length of the container string +1.*/
                dataLen = (strlen(pProvCtx->container.cName)+1)*sizeof(char);
            }
            else
            {
                dataLen = strlen(CSP_CONTAINER_NAME);
            }
            break;
        case PP_ENUMALGS: /**  - PP_ENUMALGS:*/
            /**   - Data len <= max length of the ENUMALGS structure.
             *      Currently, dword + alg_id + 16*uchar.*/
			DEBUG(3, "Handling PP_ENUMALGS\n");
            dataLen = sizeof(PROV_ENUMALGS); // WHY add + 15;
            break;

		case PP_ENUMCONTAINERS:
			if(dwFlags & CRYPT_FIRST)
            {
				dataLen = strlen(CSP_CONTAINER_NAME);
			}
			else
			{
				SetLastError(ERROR_NO_MORE_ITEMS);
                return FALSE;
			}
			break;

        case PP_KEYEXCHANGE_PIN: /**  - PP_KEYEXCHANGE_PIN:*/
            /**   - Length of the cached pin if != NULL.*/
            if(pProvCtx->cachedKeyExchangePin != NULL)
            {
                dataLen = strlen(pProvCtx->cachedKeyExchangePin);
            }
            else
            {
                dataLen = 0;
            }
            break;
        case PP_NAME: /**  - PP_NAME:*/
            /**   - Length of the csp name.*/
            dataLen = strlen(CSP_NAME);
            break;
        case PP_PROVTYPE: /**  - PP_PROVTYPE:*/
            /**   - DWORD.*/
            dataLen = sizeof(DWORD);
            break;
        case PP_SIGNATURE_PIN: /**  - PP_SIGNATURE_PIN:*/
            /**   - Length of the cached PIN if != NULL.*/
            if(pProvCtx->cachedSigPin != NULL)
            {
                dataLen = strlen(pProvCtx->cachedSigPin);
            }
            else
            {
                dataLen = 0;
            }
            break;
        case PP_VERSION: /**  - PP_VERSION:*/
            /**   - DWORD.*/
            dataLen = sizeof(DWORD);
            break;
        default:
			DEBUG(3, "Returning NTE_BAD_TYPE from GetProvParam\n");
            SetLastError(NTE_BAD_TYPE);
            return FALSE;
    }
	sprintf(debug, "Calculated required length is %d\n", dataLen);
	DEBUG(3, debug);
    /** - First call.*/
    if(pbData == NULL)
    {
		DEBUG(3, "Will set return data length\n");
        *pcbDataLen = dataLen;
		DEBUG(3, "Still alive in GetProvParam\n");
    }
    /** - Second call.*/
    else
    {

        if(*pcbDataLen < dataLen)
        {
            SetLastError(ERROR_MORE_DATA);
            return FALSE;
        }
        switch(dwParam)
        {
            case PP_CONTAINER: /**  - PP_CONTAINER:*/
                /**   - If container name is not NULL.*/
                if(pProvCtx->container.cName != NULL)
                {
                    /**   - Copy container name to pcbData.*/
                    strcpy(pbData,pProvCtx->container.cName);
                }
                else
                {
                    strcpy(pbData,CSP_CONTAINER_NAME);
                }
                break;

			case PP_ENUMCONTAINERS:
				if(dwFlags & CRYPT_FIRST)
                {
					strcpy(pbData,CSP_CONTAINER_NAME);
				}
				else
				{
					SetLastError(ERROR_NO_MORE_ITEMS);
                    return FALSE;
				}
				break;

            case PP_ENUMALGS: /**  - PP_ENUMALGS:*/
                /**   - If CRYPT_FIRST flagged.*/
                if(dwFlags & CRYPT_FIRST)
                {
                    /**    - Set current Alg to 0.*/
					DEBUG(3, "1st Algo\n");
                    pProvCtx->currentAlg = 0;
                }
                /**   - Else,*/
                else
                {
                    /**    - Next Algorithm.*/
					DEBUG(3, "Next Algo\n");
                    pProvCtx->currentAlg++;
                }
                /**   - If current Alg number >= number of supported algorithm:
                 *      end.*/
                if(pProvCtx->currentAlg>=ALG_COUNT)
                {
					DEBUG(3, "Returning ERROR_NO_MORE_ITEMS\n");
                    SetLastError(ERROR_NO_MORE_ITEMS);
                    return FALSE;
                }
                /**   - Cast pcbData pointer to local enumAlg.*/
                enumAlg = (PROV_ENUMALGS *)pbData;

                /**   - Get the algorithm information.*/
                if(!getAlgorithm(pProvCtx->currentAlg, &algorithm))
                {
                    /**     - If algorithm < 0, we are in serious trouble.*/
					DEBUG(3, "Returning NTE_FAIL...Serious Trouble\n");
                    SetLastError(NTE_FAIL);
                    return FALSE;
                }
                /**   - Fill enumAlg info.*/
                enumAlg->aiAlgid = algorithm.algId;
                enumAlg->dwBitLen = algorithm.dwBits;
                strcpy(enumAlg->szName, algorithm.cName);
                enumAlg->dwNameLen = strlen(algorithm.cName);
				DEBUG(3, "All done for ENUM_ALGS possibly...\n");
                break;

            case PP_KEYEXCHANGE_PIN: /**  - PP_KEYEXCHANGE_PIN:*/
                /**   - Cast pbData pointer to localPin.*/
                localPin = (char *)pbData;
                /**   - Copy the cached pin if !=NULL.*/
                if(pProvCtx->cachedKeyExchangePin != NULL)
                {
                    strcpy(localPin, pProvCtx->cachedKeyExchangePin);
                }
                else
                {
                    *localPin = '\0';
                }
                break;
            case PP_NAME: /**  - PP_NAME:*/
                /**   - Copy the container name.*/
                strcpy(pbData,CSP_NAME);
                break;
            case PP_PROVTYPE: /**  - PP_PROVTYPE:*/
                /**   - PROV_RSA_FULL.*/
                *pbData = PROV_RSA_FULL;
                break;
            case PP_SIGNATURE_PIN: /**  - PP_SIGNATURE_PIN:*/
                /**   - Cast pcbData pointer to localPin.*/
                localPin = (char *)pbData;
                /**   - Copy the cached pin if !=NULL.*/
                if(pProvCtx->cachedSigPin)
                {
                    strcpy(localPin, pProvCtx->cachedSigPin);
                }
                else
                {
                    *localPin = '\0';
                }
                break;
            case PP_VERSION: /**  - PP_VERSION:*/
                /**   - DWORD.*/
                *pbData = (BYTE) CSP_VERSION;
                break;
            default:
                SetLastError(NTE_BAD_TYPE);
                return FALSE;
        }
    }
    DEBUG(3, "Returning good from GetProvParam \n");
    return TRUE;
}


/*
 -  CPSetHashParam
 -
 *  Purpose:
 *                Allows applications to customize various aspects of the
 *                operations of a hash
 *
 *  Parameters:
 *               IN      hProv   -  Handle to a CSP
 *               IN      hHash   -  Handle to a hash
 *               IN      dwParam -  Parameter number
 *               IN      pbData  -  Pointer to data
 *               IN      dwFlags -  Flags values
 *
 *  Returns:
 */

BOOL WINAPI
CPSetHashParam(
    IN  HCRYPTPROV hProv,
    IN  HCRYPTHASH hHash,
    IN  DWORD dwParam,
    IN  CONST BYTE *pbData,
    IN  DWORD dwFlags)
{
	int ii;
	char debug[200];

	BOOL ret;
	PROV_CTX *pProvCtx = NULL;      /* Provider context */

	sprintf(debug, "dwParam value %d\n", dwParam);

	DEBUG(3,"------------- CPSetHashParam ----------------\n");
	DEBUG(3, debug);

    /** - Local good typed provider context pointer.*/
    pProvCtx = (PROV_CTX *) hProv;

	if(!grantedHandle((HANDLE *) grantedContexts, grantedContextsNb, (HANDLE) hProv))
    {
        SetLastError(NTE_BAD_UID);
        return FALSE;
    }

	for (ii=0; ii < 24; ii++)
	{
		sprintf(&debug[ii*2], "%02X ", pbData[ii]);
	}
	DEBUG2(3, debug, ii*2);
	DEBUG(3, "\n");

	ret = CryptSetHashParam(hHash, dwParam, pbData, dwFlags);
	if( !ret )
	{
		DEBUG(3, "MSP CPSetHashParam return bad\n");
	}
	DEBUG(3, "MSP CPSetHashParam return good\n" );
	return ret;
}

/*
 -  CPCreateHash
 -
 *  Purpose:
 *                initate the hashing of a stream of data
 *
 *
 *  Parameters:
 *               IN  hUID    -  Handle to the user identifcation
 *               IN  Algid   -  Algorithm identifier of the hash algorithm
 *                              to be used
 *               IN  hKey   -   Optional handle to a key
 *               IN  dwFlags -  Flags values
 *               OUT pHash   -  Handle to hash object
 *
 *  Returns:
 */

BOOL WINAPI
CPCreateHash(
    IN  HCRYPTPROV hProv,
    IN  ALG_ID Algid,
    IN  HCRYPTKEY hKey,
    IN  DWORD dwFlags,
    OUT HCRYPTHASH *phHash)
{
	char param[100];
	BOOL ret;
	PROV_CTX *pProvCtx = NULL;
	pProvCtx = (PROV_CTX *) hProv;

	DEBUG(3,"------------- CPCreateHash ----------------\n");

	if(!grantedHandle((HANDLE *) grantedContexts, grantedContextsNb, (HANDLE) hProv))
    {
        SetLastError(NTE_BAD_UID);
        return FALSE;
    }

	ret = CryptCreateHash(pProvCtx->msProv, Algid, hKey, dwFlags, phHash);
	if( !ret )
	{
		sprintf(param, "MS Prov Create Hash bad -> %s\n", GetLastError());
		DEBUG(3, param);
	}
	return ret;
}


/*
 -  CPHashData
 -
 *  Purpose:
 *                Compute the cryptograghic hash on a stream of data
 *
 *
 *  Parameters:
 *               IN  hProv     -  Handle to the user identifcation
 *               IN  hHash     -  Handle to hash object
 *               IN  pbData    -  Pointer to data to be hashed
 *               IN  dwDataLen -  Length of the data to be hashed
 *               IN  dwFlags   -  Flags values
 *
 *  Returns:
 */

BOOL WINAPI
CPHashData(
    IN  HCRYPTPROV hProv,
    IN  HCRYPTHASH hHash,
    IN  CONST BYTE *pbData,
    IN  DWORD cbDataLen,
    IN  DWORD dwFlags)
{
	char debug[5000];
	unsigned int ii;

	char param[100];
	BOOL ret;
	PROV_CTX *pProvCtx = NULL;
	pProvCtx = (PROV_CTX *) hProv;
	sprintf(param, "dwFlags value %d\n", dwFlags);

	DEBUG(3,"------------- CPHashData ----------------\n");
	DEBUG(3, param);

	if(!grantedHandle((HANDLE *) grantedContexts, grantedContextsNb, (HANDLE) hProv))
    {
        SetLastError(NTE_BAD_UID);
        return FALSE;
    }

	sprintf(debug, "cbDataLen value %d\n", cbDataLen);
	DEBUG(3, debug);
	for (ii=0; ii < cbDataLen; ii++)
	{
		sprintf(&debug[ii*2], "%02X ", pbData[ii]);
	}
	DEBUG2(3, debug, ii*2);
	DEBUG(3, "\n");
    ret = CryptHashData(hHash, pbData, cbDataLen, dwFlags);
	if( !ret )
	{
		DEBUG(3, " CryptHashData returned bad !!!!!\n");
	}
	DEBUG(3, " CryptHashData returned good !!!!!\n");
	return ret;
}


/*
 -  CPSignHash
 -
 *  Purpose:
 *                Create a digital signature from a hash
 *
 *
 *  Parameters:
 *               IN  hProv        -  Handle to the user identifcation
 *               IN  hHash        -  Handle to hash object
 *               IN  dwKeySpec    -  Key pair to that is used to sign with
 *               IN  sDescription -  Description of data to be signed
 *               IN  dwFlags      -  Flags values
 *               OUT pbSignature  -  Pointer to signature data
 *               IN OUT dwHashLen -  Pointer to the len of the signature data
 *
 *  Returns:
 */

BOOL WINAPI
CPSignHash(
    IN  HCRYPTPROV hProv,
    IN  HCRYPTHASH hHash,
    IN  DWORD dwKeySpec,
    IN  LPCWSTR szDescription,
    IN  DWORD dwFlags,
    OUT LPBYTE pbSignature,
    IN OUT LPDWORD pcbSigLen)
{
	int ii;
	char debug[5000];
	int signhashlen;

	PROV_CTX *pProvCtx = NULL;      /* Provider context */
    //HASH_INFO *pHash = NULL;        /* Hash information.*/
    DWORD   hashLen = 0;            /* Hash lenth in Bytes.*/
    BYTE   *pHashValue = NULL;     /* Hash value.*/
    HCRYPTKEY hKey = -1;          /* Handle to the key to use.*/
    DWORD   hashSizeLen = 4;    /* Size of an hash size data.*/
    KEY_INFO *pKeyInfo = NULL; /* Pointer to the key information structure.*/
    unsigned int i;
    /* Local function convention: hash: the raw hash, digest the asn1 encoded
     * hash.*/
    DWORD   digestLen; /* The digest length.*/
    BYTE *pDigest; /* The digest length.*/

	unsigned char tempsig[500];
	unsigned int tempsiglen = 500;

    DEBUG(3,"_-_-_-_-_-_-_-_-_Signing Hash-_-_-_-_-_-_-_-_-\n");
    /** - Local copy of the crypto handle.*/
    pProvCtx = (PROV_CTX *) hProv;

	sprintf(debug, "hProv value %x\n", hProv);
	DEBUG(3, debug);

    /** - Test if c handle has been granted.*/
    if(!grantedHandle(grantedContexts, grantedContextsNb, (HANDLE) hProv))
    {
        SetLastError(NTE_BAD_UID);
        return FALSE;
    }

	DEBUG(3, "Proceeding thru SignHash code\n");
    /**  - Get the hash lenth.*/
    if(!CryptGetHashParam(hHash, HP_HASHSIZE,(LPBYTE) &hashLen, &hashSizeLen, 0))
    {
		DEBUG(3, "Returning failure with NTE_BAD_HASH\n");
        SetLastError(NTE_BAD_HASH);
        return FALSE;
    }
    /**  - Allocate memory for the hash. Need to free - TODO*/
    pHashValue = (BYTE *) HeapAlloc(pProvCtx->heap, HEAP_ZERO_MEMORY,
                                    hashLen*sizeof(BYTE));
    if(pHashValue == NULL)
    {
		DEBUG(3, "Returning failure (1) with NTE_NO_MEMORY\n");
        SetLastError(NTE_NO_MEMORY);
        return FALSE;
    }
    /**  - Get the Hash value.*/
    if(!CryptGetHashParam(hHash, HP_HASHVAL, pHashValue, &hashLen, 0))
    {
		DEBUG(3, "Returning failure (2) with NTE_NO_MEMORY\n");
        SetLastError(NTE_BAD_HASH);
        return FALSE;
    }
	DEBUG(3, "Still going good hopefully\n");
    /** - If flags are present.*/
    if(dwFlags)
    {
		DEBUG(3, "dwFlags non zero\n");
        /**  - If no OID has to be prepend, the raw hash is signed.*/
        if(dwFlags != CRYPT_NOHASHOID)
        {
			DEBUG(3, "dwFlags non zero - BAD FLAGS FOUND!!\n");
            SetLastError(NTE_BAD_FLAGS);
            return FALSE;
        }
        DEBUG(4,"A signature without OID header has been requested.\n")
        digestLen = hashLen;
        pDigest = pHashValue;
    }
    else
    {
		DEBUG(3, "dwFlags zero\n");
        /**  - Compute the digest bytes length.*/
        digestLen = hashLen+18;
		sprintf(debug, "Hash Len = %d\n", hashLen);
		DEBUG(3, debug);
		sprintf(debug, "Digest Len = %d\n", digestLen);
		DEBUG(3, debug);

        /**  - Allocate memory for the OID and the hash.*/
        pDigest = NULL;
        pDigest = HeapAlloc(pProvCtx->heap, HEAP_ZERO_MEMORY, digestLen);
        if(pDigest == NULL)
        {
			DEBUG(3, "dwFlags zero - NO Memory\n");
            SetLastError(NTE_NO_MEMORY);
            return FALSE;
        }
		memcpy(pDigest, MD5_OID, 18);
        /**  - Copy the hash after it.*/
        for(i=0; i<hashLen; i++)
        {
            pDigest[i+18] = pHashValue[i];
        }
    }

#ifdef TLP224
	if( strlen(cachedPIN) <= 0 )
	{
		DEBUG(3, "Cached PIN not available failed\n");
        SetLastError(NTE_NO_KEY);
        return FALSE;
	}

#endif

	DEBUG(3, "Will call simpleSignData\n");
    /** - Compute the signature & sig lenth.*/
    if(!simpleSignData(pProvCtx, dwKeySpec, pDigest, digestLen, tempsig, &tempsiglen, cardHandle, cachedPIN))
    {
		DEBUG(3, "simpleSignData failed\n");
        SetLastError(NTE_BAD_ALGID);
        return FALSE;
    }
	sprintf(debug, "simpleSignData passed, Signature len->%d\n", tempsiglen);
	DEBUG(3, debug);

	// We know that we need to copy data but leave out the 1st 6 chars..Phew!!!
	*pcbSigLen = tempsiglen-6;
	if( pbSignature != NULL )
	{
		memcpy(pbSignature, &tempsig[6], tempsiglen-6);
	}

	if( pbSignature != NULL )
	{
		signhashlen = *pcbSigLen;
		for (ii=0; ii < signhashlen; ii++)
		{
			sprintf(&debug[ii*2], "%02X ", pbSignature[ii]);
		}
		DEBUG2(3, debug, ii*2);
		DEBUG(3, "\n");
	}

    /** - If the signature has been computed, reverse it.*/
    if(pbSignature != NULL)
    {
        if(!reverseBytesString(pbSignature, *((int *)pcbSigLen)))
        {
            return FALSE;
        }
    }
	DEBUG(3, "Bytes reversal done\n");
	if( pbSignature != NULL )
	{
		signhashlen = *pcbSigLen;
		for (ii=0; ii < signhashlen; ii++)
		{
			sprintf(&debug[ii*2], "%02X ", pbSignature[ii]);
		}
		DEBUG2(3, debug, ii*2);
		DEBUG(3, "\n");
	}

	DEBUG(3, "All done!!!\n");
    return TRUE;
}


/*
 -  CPDestroyHash
 -
 *  Purpose:
 *                Destroy the hash object
 *
 *
 *  Parameters:
 *               IN  hProv     -  Handle to the user identifcation
 *               IN  hHash     -  Handle to hash object
 *
 *  Returns:
 */

BOOL WINAPI
CPDestroyHash(
    IN  HCRYPTPROV hProv,
    IN  HCRYPTHASH hHash)
{
	BOOL ret;
	DEBUG(3,"------------- CPDestroyHash ----------------\n");
	/** - If context handle is granted.*/
    if(!grantedHandle((HANDLE *) grantedContexts, grantedContextsNb, (HANDLE) hProv))
    {
        SetLastError(NTE_BAD_UID);
        return FALSE;
    }
	ret = CryptDestroyHash(hHash);
	if( !ret )
	{
		DEBUG(3, " DestroyHash return bad \n!!!!!");
	}
	else
	{
		DEBUG(3, "DestroyHash success\n");
	}
	return ret;
}

/*
 -  CPGetHashParam
 -
 *  Purpose:
 *                Allows applications to get various aspects of the
 *                operations of a hash
 *
 *  Parameters:
 *               IN      hProv      -  Handle to a CSP
 *               IN      hHash      -  Handle to a hash
 *               IN      dwParam    -  Parameter number
 *               OUT     pbData     -  Pointer to data
 *               IN      pdwDataLen -  Length of parameter data
 *               IN      dwFlags    -  Flags values
 *
 *  Returns:
 */

BOOL WINAPI
CPGetHashParam(
    IN  HCRYPTPROV hProv,
    IN  HCRYPTHASH hHash,
    IN  DWORD dwParam,
    OUT LPBYTE pbData,
    IN OUT LPDWORD pcbDataLen,
    IN  DWORD dwFlags)
{
	BOOL ret;
	unsigned char debug[100];

	sprintf(debug, "dwParam value %d\n", dwParam);

	DEBUG(3,"------------- CPGetHashParam ----------------\n");
	DEBUG(3, debug);

	/** - If context handle is granted.*/
    if(!grantedHandle((HANDLE *) grantedContexts, grantedContextsNb, (HANDLE) hProv))
    {
        SetLastError(NTE_BAD_UID);
        return FALSE;
    }
	ret = CryptGetHashParam(hHash, dwParam, pbData, pcbDataLen, dwFlags);
	if( !ret )
	{
		DEBUG(3, "GetHashParam bad return\n");
	}
	/* PURELY FOR DEBUG - MAY HAVE TO DECLARE VARS!!
	if( pbData != NULL )
	{
		signhashlen = *pcbDataLen;
		for (ii=0; ii < signhashlen; ii++)
		{
			sprintf(&debug[ii*2], "%02X ", pbData[ii]);
		}
		DEBUG2(3, debug, ii*2);
		DEBUG(3, "\n");

		datalen = 5000;

		signhashlen = *pcbDataLen;
		if(Base64Encode(pbData, signhashlen, base64data, &datalen, 0))
		{
			DEBUG(3, "Base64Encode encoding fine!!!");
		}
		else
		{
			DEBUG(3, "Base64Encode encoding screwed!!!");
		}
		sprintf(debug, "Total base64 bytes returned ->%d\n", datalen);
		DEBUG(3, debug);
		for(ii=0; ii < datalen; ii++)
		{
			debug[ii] = base64data[ii];
		}
		debug[ii] = '\0';
		DEBUG(3, debug);
		sprintf(debug, "\n\n#################  $$$$$$$$$$$$$$$$$$$$$$$$$$\n");
		DEBUG(3, debug);
	}
	*/

	return TRUE;
}

/*
 -  CPGenRandom
 -
 *  Purpose:
 *                Used to fill a buffer with random bytes
 *
 *
 *  Parameters:
 *               IN  hProv         -  Handle to the user identifcation
 *               IN  dwLen         -  Number of bytes of random data requested
 *               IN OUT pbBuffer   -  Pointer to the buffer where the random
 *                                    bytes are to be placed
 *
 *  Returns:
 */

BOOL WINAPI
CPGenRandom(
    IN  HCRYPTPROV hProv,
    IN  DWORD cbLen,
    OUT LPBYTE pbBuffer)
{
	BOOL ret;
	PROV_CTX *pProvCtx = NULL;      /* Provider context */
    /** - Local good typed provider context pointer.*/
    pProvCtx = (PROV_CTX *) hProv;

	DEBUG(3,"------------- CPGenRandom ----------------\n");
    ret = CryptGenRandom(pProvCtx->msProv, cbLen, pbBuffer);
	if( !ret )
	{
		DEBUG(3, " GenRandom Returned bad retrun \n!!!!!");
	}
	DEBUG(3, " GenRandom Returned good retrun \n!!!!!");
	return ret;
}


/*
 -  CPGetUserKey
 -
 *  Purpose:
 *                Gets a handle to a permanent user key
 *
 *
 *  Parameters:
 *               IN  hProv      -  Handle to the user identifcation
 *               IN  dwKeySpec  -  Specification of the key to retrieve
 *               OUT phUserKey  -  Pointer to key handle of retrieved key
 *
 *  Returns:
 */

BOOL WINAPI
CPGetUserKey(
    IN  HCRYPTPROV hProv,
    IN  DWORD dwKeySpec,
    OUT HCRYPTKEY *phUserKey)
{
	PROV_CTX *pProvCtx = NULL;      /* Provider context */
    KEY_INFO *pKeyInfo; /* Handle to the generated key.*/
    char    debug[100]; /* Debug string.*/

    DEBUG(3,"_-_-_-_-_-_-_-_-_Getting user keys-_-_-_-_-_-_-_-_-\n");
	sprintf(debug, "dwkeyspec = %d\n", dwKeySpec);
	DEBUG(3, debug);

    /** - Nullify returned key handle.*/
    *phUserKey = 0;
    /** - Local copy of the context handle.*/
    pProvCtx = (PROV_CTX *) hProv;

    /** - Test if the context has been granted.*/
    if(!grantedHandle((HANDLE *) grantedContexts, grantedContextsNb, (HANDLE) hProv))
    {
        SetLastError(NTE_BAD_UID);
        return FALSE;
    }
    /** - Allocate memory for key information.*/
    pKeyInfo = (KEY_INFO *) HeapAlloc(pProvCtx->heap, HEAP_ZERO_MEMORY,
                                      sizeof(KEY_INFO));
    if(pKeyInfo == NULL)
    {
        SetLastError(NTE_NO_MEMORY);
        return FALSE;
    }
    /** - If requested key is not signature key, not implemented.*/
    if((dwKeySpec != AT_SIGNATURE) && (dwKeySpec != AT_KEYEXCHANGE))
    {
        SetLastError(NTE_NO_KEY);
        return FALSE;
    }
	if(dwKeySpec == AT_SIGNATURE)
	{
		// TODO - For now just consider that we are being asked for exchange - Shouvik
		//SetLastError(NTE_BAD_KEY);
		//return FALSE;
		dwKeySpec = AT_KEYEXCHANGE;
	}

    /** - Set dwKeySpec.*/
    pKeyInfo->dwKeySpec = dwKeySpec;

    /** Do the 800-73 thingy!!! */
    if(!loadUserKey(pProvCtx, pKeyInfo, cardHandle))
    {
        sprintf(debug, "Cannot get user key for usage: 0x%x\n", (unsigned int) dwKeySpec);
        DEBUG(1, debug);
        /**  - If we are not able to do it: corrupt or empty ?.*/
        SetLastError(NTE_KEYSET_ENTRY_BAD);
        return FALSE;
    }
    /**  - Fill key handle.*/
    *phUserKey = (HCRYPTKEY) pKeyInfo;
    /** - Grant key handle.*/
    DEBUG(3, "===hKey:");
    grantHandle((HANDLE **) &grantedKeys, &grantedKeysNb, (HANDLE) *phUserKey);
    return TRUE;
}


/*
 -  CPDuplicateHash
 -
 *  Purpose:
 *                Duplicates the state of a hash and returns a handle to it.
 *                This is an optional entry.  Typically it only occurs in
 *                SChannel related CSPs.
 *
 *  Parameters:
 *               IN      hUID           -  Handle to a CSP
 *               IN      hHash          -  Handle to a hash
 *               IN      pdwReserved    -  Reserved
 *               IN      dwFlags        -  Flags
 *               IN      phHash         -  Handle to the new hash
 *
 *  Returns:
 */

BOOL WINAPI
CPDuplicateHash(
    IN  HCRYPTPROV hProv,
    IN  HCRYPTHASH hHash,
    IN  LPDWORD pdwReserved,
    IN  DWORD dwFlags,
    OUT HCRYPTHASH *phHash)
{
	DEBUG(3,"------------- CPDuplicateHash ----------------\n");
	return NTE_FAIL;
}


/*
 -  CPDuplicateKey
 -
 *  Purpose:
 *                Duplicates the state of a key and returns a handle to it.
 *                This is an optional entry.  Typically it only occurs in
 *                SChannel related CSPs.
 *
 *  Parameters:
 *               IN      hUID           -  Handle to a CSP
 *               IN      hKey           -  Handle to a key
 *               IN      pdwReserved    -  Reserved
 *               IN      dwFlags        -  Flags
 *               IN      phKey          -  Handle to the new key
 *
 *  Returns:
 */

BOOL WINAPI
CPDuplicateKey(
    IN  HCRYPTPROV hProv,
    IN  HCRYPTKEY hKey,
    IN  LPDWORD pdwReserved,
    IN  DWORD dwFlags,
    OUT HCRYPTKEY *phKey)
{
	DEBUG(3,"------------- CPDuplicateKey ----------------\n");
	return NTE_FAIL;
}

/*
 -  CPGenKey
 -
 *  Purpose:
 *                Generate cryptographic keys
 *
 *
 *  Parameters:
 *               IN      hProv   -  Handle to a CSP
 *               IN      Algid   -  Algorithm identifier
 *               IN      dwFlags -  Flags values
 *               OUT     phKey   -  Handle to a generated key
 *
 *  Returns:
 */

BOOL WINAPI
CPGenKey(
    IN  HCRYPTPROV hProv,
    IN  ALG_ID Algid,
    IN  DWORD dwFlags,
    OUT HCRYPTKEY *phKey)
{
	DEBUG(3,"------------- CPGenKey ----------------\n");
	return NTE_FAIL;
}


/*
 -  CPDeriveKey
 -
 *  Purpose:
 *                Derive cryptographic keys from base data
 *
 *
 *  Parameters:
 *               IN      hProv      -  Handle to a CSP
 *               IN      Algid      -  Algorithm identifier
 *               IN      hBaseData -   Handle to base data
 *               IN      dwFlags    -  Flags values
 *               OUT     phKey      -  Handle to a generated key
 *
 *  Returns:
 */

BOOL WINAPI
CPDeriveKey(
    IN  HCRYPTPROV hProv,
    IN  ALG_ID Algid,
    IN  HCRYPTHASH hHash,
    IN  DWORD dwFlags,
    OUT HCRYPTKEY *phKey)
{
	DEBUG(3,"------------- CPDeriveKey ----------------\n");
	return NTE_FAIL;
}

/*
 -  CPSetKeyParam
 -
 *  Purpose:
 *                Allows applications to customize various aspects of the
 *                operations of a key
 *
 *  Parameters:
 *               IN      hProv   -  Handle to a CSP
 *               IN      hKey    -  Handle to a key
 *               IN      dwParam -  Parameter number
 *               IN      pbData  -  Pointer to data
 *               IN      dwFlags -  Flags values
 *
 *  Returns:
 */

BOOL WINAPI
CPSetKeyParam(
    IN  HCRYPTPROV hProv,
    IN  HCRYPTKEY hKey,
    IN  DWORD dwParam,
    IN  CONST BYTE *pbData,
    IN  DWORD dwFlags)
{
	unsigned char debug[200];
	BOOL ret;
	KEY_INFO *pKeyInfo; /* Handle to the generated key.*/

    /** - Local copy of the key handler.*/
    pKeyInfo = (KEY_INFO *) hKey;

	DEBUG(3,"------------- CPSetKeyParam ----------------\n");
	/** - Test if the context has been granted.*/
    if(!grantedHandle((HANDLE *) grantedContexts, grantedContextsNb, (HANDLE) hProv))
    {
        SetLastError(NTE_BAD_UID);
        return FALSE;
    }
	if( dwParam != KP_IV ) // We are handling only one option
	{
        SetLastError(NTE_BAD_FLAGS);
	    return FALSE;
	}
	sprintf(debug, "Incoming key handle = %x\n", hKey);
	DEBUG(3, debug);

	ret = CryptSetKeyParam(pKeyInfo->hKeyMSP, dwParam, pbData, dwFlags);
	if( !ret )
	{
		sprintf(debug, "MS Provider CryptSetKeyParam call failed = %x\n", GetLastError());
		DEBUG(3, debug);
	}
	else
	{
		DEBUG(3, "MS Provider CryptSetKeyParam call passed\n");
	}
	return ret;
}

/*
 -  CPExportKey
 -
 *  Purpose:
 *                Export cryptographic keys out of a CSP in a secure manner
 *
 *
 *  Parameters:
 *               IN  hProv         - Handle to the CSP user
 *               IN  hKey          - Handle to the key to export
 *               IN  hPubKey       - Handle to exchange public key value of
 *                                   the destination user
 *               IN  dwBlobType    - Type of key blob to be exported
 *               IN  dwFlags       - Flags values
 *               OUT pbData        -     Key blob data
 *               IN OUT pdwDataLen - Length of key blob in bytes
 *
 *  Returns:
 */

BOOL WINAPI
CPExportKey(
    IN  HCRYPTPROV hProv,
    IN  HCRYPTKEY hKey,
    IN  HCRYPTKEY hPubKey,
    IN  DWORD dwBlobType,
    IN  DWORD dwFlags,
    OUT LPBYTE pbData,
    IN OUT LPDWORD pcbDataLen)
{
	SetLastError(NTE_BAD_TYPE);
    return FALSE;
}


/*
 -  CPImportKey
 -
 *  Purpose:
 *                Import cryptographic keys
 *
 *
 *  Parameters:
 *               IN  hProv     -  Handle to the CSP user
 *               IN  pbData    -  Key blob data
 *               IN  dwDataLen -  Length of the key blob data
 *               IN  hPubKey   -  Handle to the exchange public key value of
 *                                the destination user
 *               IN  dwFlags   -  Flags values
 *               OUT phKey     -  Pointer to the handle to the key which was
 *                                Imported
 *
 *  Returns:
 */

BOOL WINAPI
CPImportKey(
    IN  HCRYPTPROV hProv,
    IN  CONST BYTE *pbData,
    IN  DWORD cbDataLen,
    IN  HCRYPTKEY hPubKey,
    IN  DWORD dwFlags,
    OUT HCRYPTKEY *phKey)
{
	HCRYPTPROV hProvLocal = 0;
	HCRYPTKEY hPubPrivKey = 0;
	HCRYPTKEY hSessionKey = 0;
	BOOL fResult;

	LPBYTE pbKeyMaterial  = NULL;

	unsigned int ii;
	unsigned char debug[1000];

	unsigned char edata[300];
	unsigned int edatalen;
	unsigned char ddata[300];
	unsigned int ddatalen = 300;
	unsigned char *sigdata = malloc(300);
	unsigned int sigdatalen;

	KEY_INFO *pKeyInfo; /* Handle to the generated key.*/
	PROV_CTX *pProvCtx = NULL;      /* Provider context */

	DEBUG(3,"------------- CPImportKey ----------------\n");

	/** - Local copy of the context handle.*/
    pProvCtx = (PROV_CTX *) hProv;

	/** - Test if the context has been granted.*/
    if(!grantedHandle((HANDLE *) grantedContexts, grantedContextsNb, (HANDLE) hProv))
    {
        SetLastError(NTE_BAD_UID);
        return FALSE;
    }

	DEBUG(3, "Grant handle check done\n");

	sprintf(debug, "hPubkey is %x\n", hPubKey);
	DEBUG(3, debug);
	sprintf(debug, "cbDataLen is %d\n", cbDataLen);
	DEBUG(3, debug);
	sprintf(debug, "dwFlags is %d\n", dwFlags);
	DEBUG(3, debug);

	memcpy(edata, pbData, cbDataLen);
	edatalen = cbDataLen;
	if(!reverseBytesString(edata, edatalen))
    {
		DEBUG(3, "Reversing string failed\n");
        return FALSE;
    }

	//Get this data decrypted by the card......
	if(!decryptData(pProvCtx, edata, edatalen, ddata, &ddatalen, cardHandle, cachedPIN))
    {
		DEBUG(3, "decryptdata failed\n");
        SetLastError(NTE_BAD_ALGID);
        return FALSE;
    }
	sprintf(debug, "decryptData passed, Signature len->%d\n", ddatalen);
	DEBUG(3, debug);

	for (ii=0; ii < ddatalen; ii++)
	{
		sprintf(&debug[ii*2], "%02X ", ddata[ii]);
	}
	DEBUG2(3, debug, ii*2);
	DEBUG(3, "\n");
	// We know that we need to copy data for 3DES....need to be made robust-Shouvik...TODO
	sigdatalen = 24;
	memcpy(sigdata, &ddata[110], sigdatalen);

	for (ii=0; ii < sigdatalen; ii++)
	{
		sprintf(&debug[ii*2], "%02X ", sigdata[ii]);
	}
	DEBUG2(3, debug, ii*2);
	DEBUG(3, "\n");

    /** - Nullify returned key handle.*/
    *phKey = 0;

    /** - Allocate memory for key information.*/
    pKeyInfo = (KEY_INFO *) HeapAlloc(pProvCtx->heap, HEAP_ZERO_MEMORY,sizeof(KEY_INFO));
    if(pKeyInfo == NULL)
    {
        SetLastError(NTE_NO_MEMORY);
        return FALSE;
    }
    //data = malloc (sizeof(PIV_Byte)*sigdatalen);
	pKeyInfo->data = sigdata;
	pKeyInfo->dataLen = sigdatalen;

    /**  - Fill key handle.*/
    *phKey = (HCRYPTKEY) pKeyInfo;
    /** - Grant key handle.*/
    DEBUG(3, "===hKey:");

	// Exponent One thingy
	__try
	{
		DEBUG(3, "Creating Exponent of One Private Key.\n\n");

		// Create Exponent of One private key
		fResult = CreateRealKey(MS_ENHANCED_PROV, PROV_RSA_FULL,
										"TestContainer", AT_KEYEXCHANGE,
										&hProvLocal, &hPubPrivKey);
		if (!fResult)
		{
			sprintf(debug, "CreateRealKey failed with %x\n", GetLastError());
			DEBUG(3, debug);
			return FALSE;
		}

		pProvCtx->msProv2 = hProvLocal;

		// Import this key and get an HCRYPTKEY handle
		if (!ImportPlainSessionBlobModified(hProvLocal, hPubPrivKey, CALG_3DES, pKeyInfo->data,
					pKeyInfo->dataLen, &hSessionKey))
		{
			sprintf(debug, "ImportPlainSessionBlobModified failed with %x\n", GetLastError());
			DEBUG(3, debug);
			return FALSE;
		}

		LocalFree(pbKeyMaterial);
		pbKeyMaterial = NULL;

		/* THE FOLLOWING WAS ONLY FOR DEVELOPMENT TIME DEBUG

			// TAKE THESE UP -- DWORD dwKeyMaterial ;unsigned int n;

		printf("Exporting 3DES key.  We should get back Original 3DES session...\n");
		// Export it and get the key material back
		// It should be 1,2,3, ...
		fResult = ExportPlainSessionBlob(hPubPrivKey, hSessionKey,
									   &pbKeyMaterial , &dwKeyMaterial );
		if (!fResult)
		{
			sprintf(debug, "ExportPlainSessionBlob failed with %x\n", GetLastError());
			DEBUG(3, debug);
			return FALSE;
		}

		// Print Key material
		DEBUG(3, "3DES session key:\n");
		for (n = 0; n < dwKeyMaterial; n++)
		{
			sprintf(&debug[ii*2], "%02X ", pbData[ii]);
		}
		DEBUG2(3, debug, ii*2);
		DEBUG(3, "\n");
		*/
	}
	__finally
	{
		if (hPubPrivKey) CryptDestroyKey(hPubPrivKey);
	}

	// Exponent One thing end
	if( !fResult )
	{
		DEBUG(3, "Exp 1 algo failed\n");
		return FALSE;
	}
	pKeyInfo->hKeyMSP = hSessionKey;
	sprintf(debug, "Key handle being returned = %x\n", pKeyInfo);
	DEBUG(3, debug);
	grantHandle((HANDLE **) &grantedKeys, &grantedKeysNb, (HANDLE) *phKey);
	return TRUE;
}


/*
 -  CPEncrypt
 -
 *  Purpose:
 *                Encrypt data
 *
 *
 *  Parameters:
 *               IN  hProv         -  Handle to the CSP user
 *               IN  hKey          -  Handle to the key
 *               IN  hHash         -  Optional handle to a hash
 *               IN  Final         -  Boolean indicating if this is the final
 *                                    block of plaintext
 *               IN  dwFlags       -  Flags values
 *               IN OUT pbData     -  Data to be encrypted
 *               IN OUT pdwDataLen -  Pointer to the length of the data to be
 *                                    encrypted
 *               IN dwBufLen       -  Size of Data buffer
 *
 *  Returns:
 */

BOOL WINAPI
CPEncrypt(
    IN  HCRYPTPROV hProv,
    IN  HCRYPTKEY hKey,
    IN  HCRYPTHASH hHash,
    IN  BOOL fFinal,
    IN  DWORD dwFlags,
    IN OUT LPBYTE pbData,
    IN OUT LPDWORD pcbDataLen,
    IN  DWORD cbBufLen)
{
	/*
    *pcbDataLen = 0;
    return TRUE;
	*/
	unsigned char debug[3000];
	BOOL ret;
	KEY_INFO *pKeyInfo; /* Handle to the generated key.*/

	DEBUG(3,"------------- CPDecrypt ----------------\n");
	/** - Test if the context has been granted.*/
    if(!grantedHandle((HANDLE *) grantedContexts, grantedContextsNb, (HANDLE) hProv))
    {
        SetLastError(NTE_BAD_UID);
        return FALSE;
    }

	sprintf(debug, "Data length being passed in = %d\n", *pcbDataLen);
	DEBUG(3, debug);

	sprintf(debug, "dwFlags passed in = %d\n", dwFlags);
	DEBUG(3, debug);

	sprintf(debug, "fFinal Flag is = %d\n", fFinal);
	DEBUG(3, debug);

	sprintf(debug, "Hash handle being passed in = %x\n", hHash);
	DEBUG(3, debug);

	sprintf(debug, "Key handle being passed in = %x\n", hKey);
	DEBUG(3, debug);

    /** - Local copy of the key handler.*/
    pKeyInfo = (KEY_INFO *) hKey;
	ret = CryptEncrypt(pKeyInfo->hKeyMSP, hHash, fFinal, dwFlags, pbData, pcbDataLen, cbBufLen);
	if( !ret )
	{
		sprintf(debug, "MSP CryptEncrypt failed = %s\n", GetLastError());
		DEBUG(3, debug);
		return FALSE;
	}
	DEBUG(3, "MSP CryptEncrypt passed\n");
	return TRUE;
}


/*
 -  CPDecrypt
 -
 *  Purpose:
 *                Decrypt data
 *
 *
 *  Parameters:
 *               IN  hProv         -  Handle to the CSP user
 *               IN  hKey          -  Handle to the key
 *               IN  hHash         -  Optional handle to a hash
 *               IN  Final         -  Boolean indicating if this is the final
 *                                    block of ciphertext
 *               IN  dwFlags       -  Flags values
 *               IN OUT pbData     -  Data to be decrypted
 *               IN OUT pdwDataLen -  Pointer to the length of the data to be
 *                                    decrypted
 *
 *  Returns:
 */

BOOL WINAPI
CPDecrypt(
    IN  HCRYPTPROV hProv,
    IN  HCRYPTKEY hKey,
    IN  HCRYPTHASH hHash,
    IN  BOOL fFinal,
    IN  DWORD dwFlags,
    IN OUT LPBYTE pbData,
    IN OUT LPDWORD pcbDataLen)
{
	unsigned char debug[3000];
	BOOL ret;

	KEY_INFO *pKeyInfo; /* Handle to the generated key.*/

	DEBUG(3,"------------- CPDecrypt ----------------\n");
	/** - Test if the context has been granted.*/
    if(!grantedHandle((HANDLE *) grantedContexts, grantedContextsNb, (HANDLE) hProv))
    {
        SetLastError(NTE_BAD_UID);
        return FALSE;
    }

	sprintf(debug, "Data length being passed in = %d\n", *pcbDataLen);
	DEBUG(3, debug);
	sprintf(debug, "dwFlags passed in = %d\n", dwFlags);
	DEBUG(3, debug);
	sprintf(debug, "fFinal Flag is = %d\n", fFinal);
	DEBUG(3, debug);
	sprintf(debug, "Hash handle being passed in = %x\n", hHash);
	DEBUG(3, debug);
	sprintf(debug, "Key handle being passed in = %x\n", hKey);
	DEBUG(3, debug);

    /** - Local copy of the key handler.*/
    pKeyInfo = (KEY_INFO *) hKey;
	ret = CryptDecrypt(pKeyInfo->hKeyMSP, hHash, fFinal, dwFlags, pbData, pcbDataLen);
	if( !ret )
	{
		sprintf(debug, "MSP CryptDecrypt failed = %s\n", GetLastError());
		DEBUG(3, debug);
		return FALSE;
	}
	DEBUG(3, "MSP CryptDecrypt passed\n");
	return TRUE;
}

/*
 -  CPHashSessionKey
 -
 *  Purpose:
 *                Compute the cryptograghic hash on a key object.
 *
 *
 *  Parameters:
 *               IN  hProv     -  Handle to the user identifcation
 *               IN  hHash     -  Handle to hash object
 *               IN  hKey      -  Handle to a key object
 *               IN  dwFlags   -  Flags values
 *
 *  Returns:
 *               CRYPT_FAILED
 *               CRYPT_SUCCEED
 */

BOOL WINAPI
CPHashSessionKey(
    IN  HCRYPTPROV hProv,
    IN  HCRYPTHASH hHash,
    IN  HCRYPTKEY hKey,
    IN  DWORD dwFlags)
{
	DEBUG(3,"------------- CPHashSessionKey ----------------\n");
	return NTE_FAIL;
}

/*
 -  CPVerifySignature
 -
 *  Purpose:
 *                Used to verify a signature against a hash object
 *
 *
 *  Parameters:
 *               IN  hProv        -  Handle to the user identifcation
 *               IN  hHash        -  Handle to hash object
 *               IN  pbSignture   -  Pointer to signature data
 *               IN  dwSigLen     -  Length of the signature data
 *               IN  hPubKey      -  Handle to the public key for verifying
 *                                   the signature
 *               IN  sDescription -  String describing the signed data
 *               IN  dwFlags      -  Flags values
 *
 *  Returns:
 */

BOOL WINAPI
CPVerifySignature(
    IN  HCRYPTPROV hProv,
    IN  HCRYPTHASH hHash,
    IN  CONST BYTE *pbSignature,
    IN  DWORD cbSigLen,
    IN  HCRYPTKEY hPubKey,
    IN  LPCWSTR szDescription,
    IN  DWORD dwFlags)
{
	DEBUG(3,"------------- CPVerifySignature ----------------\n");
	return NTE_FAIL;
}